#include <pthread.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

/* number of attempts for an address */
#define PING_RETRIES 3
/* timeout for echo reply (in ms) */
#define PING_TIMEOUT 3000


typedef struct {
  pthread_t id;
  const char* ip;
  int status;
} child_info;


u_int16_t icmp_ident;


inline u_int64_t now()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return 1000000UL*tv.tv_sec + tv.tv_usec;
}


inline u_int16_t calc_sum(void* buff, size_t size)
{
  int i;
  u_int32_t S=0;
  u_int16_t* p = buff;
  size = size / 2;
  
  for (i=0; i<size; i++) S += *p++;
  return ~((S & 0xFFFF) + (S >> 16));
}


void* child_main(void* _info)
{
  child_info* info = (child_info*) _info;
  int sd;
  int R;
  u_int16_t i;
  ssize_t L;
  struct sockaddr_in target;
  u_int32_t target_addr;
  socklen_t target_len;
  struct icmphdr icmp_out;
  char buff[128];
  struct pollfd pollinfo;
  u_int64_t start;
  unsigned long timeout, elapsed;
  
  target.sin_family = AF_INET;
  target.sin_port = 0;
  R = inet_pton(AF_INET, info->ip, &target.sin_addr);
  if (R < 0) return NULL;
  target_addr = target.sin_addr.s_addr;
  icmp_out.type = ICMP_ECHO;
  icmp_out.code = 0;
  icmp_out.un.echo.id = icmp_ident; 
  
  sd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (sd < 0) return NULL;
  pollinfo.fd = sd;
  pollinfo.events = POLLIN;
  pollinfo.revents = 0;
  
  for (i=0; i<PING_RETRIES; i++) {
    icmp_out.un.echo.sequence = htons(i);
    icmp_out.checksum = 0;
    icmp_out.checksum = calc_sum(&icmp_out, sizeof(icmp_out));
    L = sendto(sd, &icmp_out, sizeof(icmp_out), 0,
               (struct sockaddr*) &target, sizeof(target));
    if (L < 0) continue;
    
    start = now();
    timeout = PING_TIMEOUT;
    while (timeout > 0) {
      R = poll(&pollinfo, 1, timeout);
      if (R<0 && errno!=EINTR) return NULL;
      if (R > 0) {
        target_len = sizeof(target);
        L = recvfrom(sd, buff, sizeof(buff), 0,
                    (struct sockaddr*) &target, &target_len);
        if (L > 0) {
          struct iphdr* ip_in = (struct iphdr*) buff;
          fflush(stdout);
          if (ip_in->saddr == target_addr) {
            struct icmphdr* icmp_in = (struct icmphdr*) (buff + 4*ip_in->ihl);
            if (icmp_in->type==ICMP_ECHOREPLY && icmp_in->code==0
                && icmp_in->un.echo.id==icmp_ident
                && icmp_in->un.echo.sequence==i) {
              info->status = 1;
              return(NULL); 
            }
          }
        }
      }
      elapsed = (now() - start) / 1000;
      if (elapsed > PING_TIMEOUT) break;
      timeout = PING_TIMEOUT - elapsed;
    }
  }
  
  return NULL;
}


int main(int argc, char* argv[])
{
  child_info* children;
  int R;
  int i;
  
  if (geteuid()) {
    fprintf(stderr, "%s: must be run as root\n", argv[0]);
    return 3;
  }
  icmp_ident = (u_int16_t) getpid();
  children = (child_info*) malloc(argc * sizeof(child_info));
  if (!children) return 1;

  for (i=1; i<argc; i++) {
    children[i].ip = argv[i];
    children[i].status = 0;
    R = pthread_create(&children[i].id, NULL, child_main, &children[i]);
    if (R) return(2);
  }
  
  for (i=1; i<argc; i++)
    pthread_join(children[i].id, NULL);
  for (i=1; i<argc; i++)
    printf("%-16s%d\n", children[i].ip, children[i].status);

  return 0;
}
